
/*
 * r1bindex.c -- right plus 1 or beginning index
 *
 * $Id$
 */

#include <h/mh.h>


char *
r1bindex(char *str, int chr)
{
    register char *cp;

    for (cp = str; *cp; cp++)
	continue;
    --cp;
    while (cp >= str && *cp != chr)
	--cp;
    return (++cp);
}
