
/*
 * m_seqflag.c -- return the flag (bit) for a sequence
 *
 * $Id$
 */

#include <h/mh.h>


int
m_seqflag (struct msgs *mp, char *cp)
{
    int i;

    for (i = 0; mp->msgattrs[i]; i++)
	if (!strcmp (mp->msgattrs[i], cp))
	    return (1 << (FFATTRSLOT + i));

    return 0;
}
