
/*
 * m_seqbits.c -- return the sprintb() string for a sequence
 *
 * $Id$
 */

#include <h/mh.h>


char *
m_seqbits (struct msgs *mp)
{
    int i;
    static char buffer[BUFSIZ];

    strcpy (buffer, MBITS);
    for (i = 0; mp->msgattrs[i]; i++)
	sprintf (buffer + strlen (buffer), "%c%s",
		FFATTRSLOT + 1 + i, mp->msgattrs[i]);

    return buffer;
}
