
/*
 * remdir.c -- remove a directory
 *
 * $Id$
 */

#include <h/mh.h>


int
remdir (char *dir)
{
    m_update();
    fflush(stdout);

    if (rmdir(dir) == NOTOK) {
	admonish (dir, "unable to remove directory");
	return 0;
    }
    return 1;
}
