
/*
 * print_help.c -- print a help message
 *
 * $Id$
 */

#include <h/mh.h>


void
print_help (char *str, struct swit *swp)
{
    char *s;

    printf ("Usage: %s\n", str);
    printf ("  switches are:\n");
    print_sw (ALL, swp, "-");
    if ((s = m_find(invo_name)))
	printf ("\nProfile: %s\n", s);
}
