
/*
 * m_getfolder.c -- get the current folder
 *
 * $Id$
 */

#include <h/mh.h>


char *
m_getfolder(void)
{
    register char *folder;

    if ((folder = m_find (pfolder)) == NULL || *folder == 0)
	folder = defalt;
    return folder;
}
