
/*
 * ssequal.c -- initially equal?
 *
 * $Id$
 */

#include <h/mh.h>


int
ssequal (char *substr, char *str)
{
    if (!substr)
	substr = "";
    if (!str)
	str = "";

    while (*substr)
	if (*substr++ != *str++)
	    return 0;
    return 1;
}
