
/*
 * closefds.c -- close-up fd's
 *
 * $Id$
 */

#include <h/mh.h>


void
closefds(int i)
{
    int nbits = OPEN_MAX;

    for (; i < nbits; i++)
#ifdef	OVERHEAD
	if (i != fd_def && i != fd_ctx)
#endif	/* OVERHEAD */
	    close (i);
}
